/* Copyright (c) 1996, 1997, 1998, 1999 by Solid Data Systems.
   *
   * Permission to use, copy, modify, and distribute this software for any
   * purpose with or without fee is hereby granted, provided that the above
   * copyright notice and this permission notice appear in all copies.
   *
   * THE SOFTWARE IS PROVIDED "AS IS" AND SOLID DATA SYSTEMS DISCLAIMS
   * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED 
   * WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL SOLID DATA
   * SYSTEMS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
   * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
   * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
   * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
   * THIS SOFTWARE.
 */

/******************************************************/
/*   maint.c    various maintenance functions V2.30   */
/*   INPUT:                                           */
/*   int fd        file descriptor for device to test */
/*   long blocks    # of blocks on the device to test */
/*   Richard Cooper  11/19/97                         */
/******************************************************/

/*************************************************************/
/*  Fixed a problem in the reliability test.  Transfer sizes */
/*  were being incremented by 512 bytes after each pass when */
/*  a transfer size less then 65536 was requested            */
/*  rbc 10/11/99                                             */
/*************************************************************/

#include "IOtest.h"
#include <termios.h>		/* get structure for termios */

extern void timeofday(void);
extern void current_time(void);
extern void current_time_to_log(void);
static unsigned long int swap(unsigned long int data_to_swap);
static void dump(int fd, long long block_number_to_dump, int display_mode);

static unsigned long wbuffer[MAX_TRANSFER_SIZE / 4];
static unsigned long rbuffer[MAX_TRANSFER_SIZE / 4];
extern long long blocks;
extern int fd;
extern int do_log, skip_block_zero, min_xfer_size, max_xfer_size;
int big_endian;
static int y = 1;
extern char rawdevice[20];

/************************************************/
/* Reliability  (Test 1)                        */
/*                                              */
/* Write zeros, read back, and compare.  Repeat */
/* with one's, aaaa's, and 5555's               */
/* transfer size = max_xfer_size                */
/* Set _ALL_SIZES = 1 for all transfer sizes    */
/************************************************/

void reliability(int fd, long long blocks)
{
    long charswrote, charsread;
    long transfer_size = 0;
    int pass_count;
    int j;
    int i;
    int all_transfer_sizes = 0;
    int min_transfer = 0, max_transfer = 0;
    unsigned long data[] =
    {0x00000000, 0xFFFFFFFF, 0xaaaaaaaa, 0x55555555, 0x0F0F0F0F, 0xF0F0F0f0, 0xFF00FF00, 0x00FF00FF};
    int skip_zero = skip_block_zero;
    int PATTERNS;
    long long start_block;	/* starting block # for transfer  */

#ifdef PERCENT
    float percent = 0;
    int percent_incr = 5;
#endif

#ifdef _ALL_SIZES
    all_transfer_sizes = 1;
#endif

    printf("Starting reliability test on device %s at ", rawdevice);
    timeofday();
    if (do_log == 1) {
	sprintf(buf, "Starting reliability test on device %s at ", rawdevice);
	logit(buf);
	current_time_to_log();
    }
    PATTERNS = sizeof(data) / 4;
    if (skip_block_zero == 1) {
	printf("Block zero(0) will not be tested\n");
    }
    for (pass_count = 1; pass_count <= REL_PASS_COUNT; ++pass_count) {

	/*transfer_size = MAX_TRANSFER_SIZE;*/
	transfer_size = max_xfer_size;
	min_transfer = min_xfer_size;
	max_transfer = max_xfer_size;

	if (all_transfer_sizes == 1) {
	    printf("Transfer sizes from %d to %d bytes will be used\n", min_transfer, max_transfer);
	} else
	    min_transfer = max_transfer;
	for (transfer_size = min_transfer; transfer_size <= max_xfer_size; transfer_size += 512) {
	    printf("\nTransfer size = %ld bytes", transfer_size);

	    for (j = 0; j < PATTERNS; ++j) {	/* fill write buffer  */

#ifdef PERCENT
		percent = 0;
#endif

		for (i = 0; i < (int) (transfer_size / 4); ++i) {
		    wbuffer[i] = data[j];
		}
/****************/
/* Writing ...  */
/****************/
		/*printf("\nwriting with %08lx's       ", data[j]); */

		sprintf(buf, "\nwriting with %08lx's       ", data[j]);
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);

		start_block = (long long) skip_zero;
		while (start_block < blocks) {

#ifdef PERCENT
		    if ((float) ((float) start_block / (float) blocks * 100) >= percent) {
			printf("\b\b\b\b%3.0f%%", (float) ((float) start_block / (float) blocks * 100));
			percent = percent + percent_incr;
			if ((int) percent == 100)
			    printf("\b\b\b\b100%%");
			if (fflush(stdout) == EOF) {
			    printf("Error flushing stdout\n");
			    exit(EXIT_FAILURE);
			}
		    }
#endif
#ifdef _LARGEFILE64_SOURCE
		    if ((lseek64(fd, (off64_t) (start_block * 512), SEEK_SET)) < 0) {
			perror("IOtest: lseek64");
			exit(EXIT_FAILURE);
		    }
#else
		    if ((lseek(fd, (start_block * 512), SEEK_SET)) < 0) {
			perror("IOtest: lseek");
			exit(EXIT_FAILURE);
		    }
#endif
		    if ((charswrote = (long int) write(fd, wbuffer, (size_t) transfer_size)) < 0) {
			perror("IOtest: write");
			exit(EXIT_FAILURE);
		    }
		    start_block = start_block + (transfer_size / 512);
		}
/*************************************************************/
/*  Reading and Comparing                                    */
/*************************************************************/
#ifdef PERCENT
		percent = 0;
#endif
		start_block = (long long) skip_zero;
		sprintf(buf, "\nreading and comparing %08lx's        ", data[j]);
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		while (start_block < blocks) {
#ifdef PERCENT
		    if ((float) ((float) start_block / (float) blocks * 100) >= percent) {
			printf("\b\b\b\b%3.0f%%", (float) ((float) start_block / (float) blocks * 100));
			percent = percent + percent_incr;
			if ((int) percent == 100)
			    printf("\b\b\b\b100%%");
			if (fflush(stdout) == EOF) {
			    printf("Error flushing stdout\n");
			    exit(EXIT_FAILURE);
			}
		    }
#endif
#ifdef _LARGEFILE64_SOURCE
		    if ((lseek64(fd, (start_block * 512), SEEK_SET)) < 0) {
			perror("IOtest: lseek64");
			exit(EXIT_FAILURE);
		    }
#else
		    if ((lseek(fd, (start_block * 512), SEEK_SET)) < 0) {
			perror("IOtest: lseek");
			exit(EXIT_FAILURE);
		    }
#endif
		    if ((charsread = (long int) read(fd, rbuffer, (size_t) transfer_size)) < 0) {
			perror("IOtest: read");
			exit(EXIT_FAILURE);
		    }
		    /*  data read back is compared to the write buffer using "!="         */
		    /*  This should be done differently, keeping track of bit positions.  */

		    for (i = 0; i < (int) (transfer_size / 4); ++i) {
			if (wbuffer[i] != rbuffer[i]) {

			    sprintf(buf, "\nError comparing write and read buffers\n");
			    printf("%s", buf);
			    if (do_log == 1)
				logit(buf);
			    sprintf(buf, "block = %lld\n", start_block + (i / 128));
			    printf("%s", buf);
			    if (do_log == 1)
				logit(buf);
			    sprintf(buf, "Error is in 32 bits starting at byte %d\n", (i % 128) * 4 + 1);
			    printf("%s", buf);
			    if (do_log == 1)
				logit(buf);
			    sprintf(buf, "Expected data = %08lX  Received data = %08lX\n", swap(wbuffer[i]), swap(rbuffer[i]));
			    printf("%s", buf);
			    if (do_log == 1)
				logit(buf);
			    sprintf(buf, "     byte %3d --^^             byte %3d --^^\n", i % 128 * 4 + 1, i % 128 * 4 + 1);
			    printf("%s", buf);
			    if (do_log == 1)
				logit(buf);
			    sprintf(buf, "\nHEX dump of block # %lld\n", start_block + (i / 128));
			    printf("%s", buf);
			    if (do_log == 1)
				logit(buf);
			    dump(fd, start_block + (i / 128), 1);
			    exit(EXIT_FAILURE);
			}
		    }

		    start_block = start_block + (transfer_size / 512);
		}
	    }

	}
	sprintf(buf, "\nEnd of pass %d at ", pass_count);
	printf("%s", buf);
	timeofday();
	if (do_log == 1)
	    logit(buf);
	current_time_to_log();
    }
}

/****************************************/
/* Quick read  (Test 2)                 */
/*                                      */
/* Read device using MAX_TRANSFER_SIZE  */
/* transfers                            */
/****************************************/
void quick_read(int fd, long long blocks)
{
    long long start_block = 0;
    long charsread;
    char rbuff[MAX_TRANSFER_SIZE];
#ifdef PERCENT
    float percent = 0;
    int percent_incr = 5;
#endif
    printf("Starting quick read at ");
    timeofday();
    if (do_log == 1) {
	sprintf(buf, "Starting quick read at ");
	logit(buf);
	current_time_to_log();
    }
    while (start_block < (blocks - 1)) {
#ifdef PERCENT
	if ((float) ((float) start_block / (float) blocks * 100) >= percent) {
	    printf("\b\b\b\b%3.0f%%", (float) ((float) start_block / (float) blocks * 100));
	    percent = percent + percent_incr;
	    if ((int) percent == 100)
		printf("\b\b\b\b100%%");
	    if (fflush(stdout) == EOF) {
		printf("Error flushing stdout\n");
		exit(EXIT_FAILURE);
	    }
	}
#endif
#ifdef _LARGEFILE64_SOURCE
	if ((lseek64(fd, (off64_t) (start_block * 512), SEEK_SET)) < 0) {
	    printf(buf, "lseek64 error\n");
	    printf("%s", buf);
	    if (do_log == 1)
		logit(buf);
	    exit(EXIT_FAILURE);
	}
#else
	if ((lseek(fd, (start_block * 512), SEEK_SET)) < 0) {
	    printf(buf, "lseek error\n");
	    printf("%s", buf);
	    if (do_log == 1)
		logit(buf);
	    exit(EXIT_FAILURE);
	}
#endif
	if ((charsread = (long int) read(fd, rbuff, (size_t) MAX_TRANSFER_SIZE)) < 0) {
	    io_error("read error");
	}
	if (charsread < MAX_TRANSFER_SIZE) {

	    printf(buf, "\nDone!  %lld blocks read \n", start_block + (charsread / 512));
	    printf("%s", buf);
	    if (do_log == 1)
		logit(buf);
	    return;
	}
	start_block = start_block + (MAX_TRANSFER_SIZE / 512);
    }
    printf("\nDone ! %lld blocks read at ", start_block);
    timeofday();
    if (do_log == 1) {
	sprintf(buf, "Done! %lld blocks read at ", start_block);
	logit(buf);
	current_time_to_log();
    }
}

/********************************************/
/* Write Blocks with block numbers (test 3) */
/********************************************/
void write_block(int fd, long long blocks)
{
#ifdef PERCENT
    float percent = 0;
    int percent_incr = 5;
#endif
    int charswrote;
    int i = 0, j = 0, k = 0;
    long long start_block;
    /* long long where_it_went = 0; */
    printf("\nStarting write block at ");
    timeofday();

    if (do_log == 1) {
	sprintf(buf, "Starting write block at ");
	logit(buf);
	current_time_to_log();
    }
    start_block = 0;

    while (start_block < blocks) {

#ifdef PERCENT
	if ((float) ((float) start_block / (float) blocks * 100) >= percent) {
	    printf("\b\b\b\b%3.0f%%", (float) ((float) start_block / (float) blocks * 100));
	    percent = percent + percent_incr;
	    if ((int) percent == 100)
		printf("\b\b\b\b100%%");
	    if (fflush(stdout) == EOF) {
		printf("Error flushing stdout\n");
		exit(EXIT_FAILURE);
	    }
	}
#endif

/****************************************************************/
/* I'll document the next few lines later.  The process of      */
/* filling the wbuffer with the block numbers before writing    */
/* it out to the disk is fairly involved; and somewhat cryptic. */
/****************************************************************/
	k = 0;
	for (i = 0; i < (MAX_TRANSFER_SIZE / 4); i = i + 128) {
	    for (j = 0; j < 128; ++j) {
		wbuffer[i + j] = (unsigned long int) start_block + k;
		/*printf("wbuffer[%d] = %ld\n",j,wbuffer[i]); */
	    }
	    k = k + 1;
	}
#ifdef _LARGEFILE64_SOURCE
	if ((lseek64(fd, (off64_t) (start_block * 512), SEEK_SET)) < 0) {
	    perror("IOtest: lseek64");
	    exit(EXIT_FAILURE);
	}
#else
	if ((lseek(fd, (start_block * 512), SEEK_SET)) < 0) {
	    perror("IOtest: lseek");
	    exit(EXIT_FAILURE);
	}
#endif
	if ((charswrote = (int) write(fd, wbuffer, (size_t) MAX_TRANSFER_SIZE)) < 0) {
	    perror("IOtest: write");
	    exit(EXIT_FAILURE);
	}
	start_block = start_block + (MAX_TRANSFER_SIZE / 512);
    }
    printf("\nDone!  %lld blocks written at ", blocks);
    timeofday();
    if (do_log == 1) {
	sprintf(buf, "Done!  Write block finished at ");
	logit(buf);
	current_time_to_log();
    }
}

/***********************************/
/* Scan block numbers (Test 4)     */
/***********************************/
void scan_blocks(int fd, long long blocks)
{
#ifdef PERCENT
    float percent = 0;
    int percent_incr = 5;
#endif
    int i = 0, j = 0, k = 0;
    int charsread;
    long long start_block;

    start_block = 0;
    printf("Starting block number scan at ");
    timeofday();
    if (do_log == 1) {
	sprintf(buf, "Starting block number scan at ");
	logit(buf);
	current_time_to_log();
    }
    while (start_block < blocks) {

#ifdef PERCENT
	if ((float) ((float) start_block / (float) blocks * 100) >= percent) {
	    printf("\b\b\b\b%3.0f%%", (float) ((float) start_block / (float) blocks * 100));
	    percent = percent + percent_incr;
	    if ((int) percent == 100)
		printf("\b\b\b\b100%%");
	    if (fflush(stdout) == EOF) {
		printf("Error flushing stdout\n");
		exit(EXIT_FAILURE);
	    }
	}
#endif

	k = 0;
	/* fill write buffer with expected data   */
	for (i = 0; i < (MAX_TRANSFER_SIZE / 4); i = i + 128) {
	    for (j = 0; j < 128; ++j) {
		wbuffer[i + j] = (unsigned long int) start_block + k;
	    }
	    k = k + 1;
	}
	/* Go get data from device and put it in rbuffer  */
#ifdef _LARGEFILE64_SOURCE
	if ((lseek64(fd, (off64_t) (start_block * 512), SEEK_SET)) < 0) {
	    perror("IOtest: lseek64");
	    exit(EXIT_FAILURE);
	    exit(EXIT_FAILURE);
	}
#else
	if ((lseek(fd, (start_block * 512), SEEK_SET)) < 0) {
	    perror("IOtest: lseek");
	    exit(EXIT_FAILURE);
	}
#endif
	if ((charsread = (int) read(fd, rbuffer, (size_t) MAX_TRANSFER_SIZE)) < 0) {
	    perror("IOtest: read");
	    exit(EXIT_FAILURE);
	}
	if (charsread < MAX_TRANSFER_SIZE) {

	    printf("\nDone!  %lld blocks read at ", start_block + (charsread / 512));
	    timeofday();
	    if (do_log == 1)
		sprintf(buf, "Done! %lld blocks read at ", start_block + (charsread / 512));
	    logit(buf);
	    current_time_to_log();
	    return;
	}
	for (i = 0; i < (MAX_TRANSFER_SIZE / 4); ++i) {
	    /* compare wbuffer and rbuffer                   */
	    if (wbuffer[i] != rbuffer[i]) {
		sprintf(buf, "\nError comparing expected and received data at ");
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		sprintf(buf, "block %lld\n", start_block + (i / 128));
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		sprintf(buf, "Error is in 32 bits starting at byte %d\n", (i % 128) * 4 + 1);
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		sprintf(buf, "Expected data = %08lX  Received data = %08lX\n", swap(wbuffer[i]), swap(rbuffer[i]));
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		sprintf(buf, "     byte %3d --^^             byte %3d --^^\n", i % 128 * 4 + 1, i % 128 * 4 + 1);
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		sprintf(buf, "HEX block dump of block %lld\n", start_block + (i / 128));
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
		dump(fd, start_block + (i / 128), 1);
		exit(EXIT_FAILURE);
	    }
	}
/*  Increment starting block # and do it again   */
	start_block = start_block + (MAX_TRANSFER_SIZE / 512);
    }

    printf("\nDone!  %lld blocks scanned at ", blocks);
    timeofday();
    if (do_log == 1) {
	sprintf(buf, "Done! block number scan finished at ");
	logit(buf);
	current_time_to_log();
    }
}


/************************************************/
/*  Test 5  continuous write block / scan block */
/************************************************/
void write_scan_blocks(int fd, long long blocks)
{
    while (1) {
	write_block(fd, blocks);
	scan_blocks(fd, blocks);
    }
}

/*******************************************************************/
/*  Dump block number (Test 6)                                     */
/*******************************************************************/
void dump_block(int fd, long long blocks)
{
    int i, j;
    long long block_number_to_dump = 0;
    char line[10];
    char character;
    static int display_mode = 1;
/*unsigned long int testword={0x01020304}; */
/******************************************************************/
/*  The next 6 lines are really unnecessary;  we always get input */
/*  from stdin (0)                                                */
/******************************************************************/
/*FILE  *input; */
/*input = fopen("/dev/tty", "r"); open tty read-only  */
/*if(!input) { */
/*      fprintf(stderr, "Unable to open /dev/tty for input\n"); */
/*      exit(EXIT_FAILURE); */
/*} */

/*******************************************************************/
/* The following lines save the terminal settings and then put the */
/* terminal in "raw" mode.  Characters are passed to the program   */
/* immediately without waiting for a "\n" (unbuffered mode)        */
/* This allows you to press "n" to get next block without a return */
/*******************************************************************/
    static struct termios initial_settings;
    static struct termios new_settings;

    for (j = 0; j <= 10; ++j)	/* zero out line buffer  */
	line[j] = (char) 0x0;

    if ((tcgetattr(0, &initial_settings)) < 0) {
	printf("tcgetattr error\n");
	exit(EXIT_FAILURE);
    }
    new_settings = initial_settings;
    new_settings.c_lflag &= ~ICANON;	/*  turn off canonical mode   */
    new_settings.c_cc[VMIN] = (cc_t) 1;
    new_settings.c_cc[VTIME] = (cc_t) 0;
    new_settings.c_lflag &= ~ISIG;

    if (tcsetattr(0, TCSANOW, &new_settings) != 0) {
	fprintf(stderr, "could not set attributes\n");
    }
/*  Print header and go get block number to test  */
/*  Do a couple of sanity checks                  */
/*****************************************************************/
/*  The next line gets one character from stdin; and ignores it  */
/*  For some reason stdin has a character "waiting".  If you     */
/*  don't do this, the main loop sees a request for block "0"    */
/*  the first time through.                                      */
/*****************************************************************/
    /* character = (char) getchar(); */

    printf("\nHEX block dump  \"q\" to quit\n");
    printf("next block = \"n\"  previous block = \"p\"\n");
    printf("Display in HEX = \"h\", ASCII = \"a\"\n");

    while ((int) block_number_to_dump >= -1) {
	printf("Enter block number (0 to %lld):  ", blocks - 1);

	i = 0;			/* reset character counter for next block # request */
	while (1) {
	    character = (char) getchar();	/* get character from stdin */
	    if (character == 'n') {	/* next block?? */
		if (block_number_to_dump < (blocks - 1))
		    block_number_to_dump++;
		break;
	    }
	    if (character == 'p') {	/* previous block?? */
		if (block_number_to_dump >= 1)
		    block_number_to_dump--;
		break;
	    }
	    if (character == 'h') {	/* Display in HEX???? */
		display_mode = 1;
		break;
	    }
	    if (character == 'a') {	/* Display in ascii?? */
		display_mode = 0;
		break;
	    }
	    if (character == 'q') {	/* quit?? */
		if ((tcsetattr(0, TCSANOW, &initial_settings)) < 0) {
		    printf("tssetattr error\n");
		    exit(EXIT_FAILURE);
		}
		return;
	    }
	    if (character == '\n') {	/* some one hit return */
		block_number_to_dump = (unsigned int) atoi(line);
		for (j = 0; j <= 10; ++j)
		    line[j] = (char) 0x0;	/*zero buffer */
		break;
	    }
	    line[i] = character;	/* one character of a block # */
	    ++i;		/* ++i, and go get next character */
	}

/*  We have a block #!! Check it against size of device  */

	if (block_number_to_dump >= blocks) {
	    printf("Block number too large for this device. \n");
	    if ((tcsetattr(0, TCSANOW, &initial_settings)) < 0) {
		printf("tcsetattr error\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	}
	if (block_number_to_dump < 0) {		/* negative block # - exit */
	    if ((tcsetattr(0, TCSANOW, &initial_settings)) < 0) {	/* reset terminal */
		printf("tcsetattr error\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	}
/*  Valid block #  dump it to screen  */

	sprintf(buf, "\nBlock number %lld\n", block_number_to_dump);
	printf("%s", buf);
	if (do_log == 1) {
	    logit(buf);
	}
	dump(fd, block_number_to_dump, display_mode);	/* dump the block to the screen in HEX */
    }
}

/********************************************************************/
/*  Test 7                                                          */
/* Write a byte. This function will prompt for a block #, byte #,   */
/* and one byte of data.  The block number is read, the data at the */
/* byte location is changed, and the block is written back to the   */
/* device.                                                          */
/* WARNING WARNING!! There are no limits to the amount of damage    */
/* you can cause with this test!!!                                  */
/********************************************************************/
void write_a_byte(int fd, long long blocks)
{
    int byte = 1, i = 0, j = 0;
    static int data_byte;
    int charsread, charswrote;
    long long block_number_to_dump = 0;
    char line[5];
    char character;
    char wbuff[512];
    static int display_mode = 1;
    for (j = 0; j <= 5; ++j)	/*  fill buffer with 0's  */
	line[j] = (char) 0x0;

    printf("Write a byte // Enter negative block number to quit.\n");
    while (1) {
	i = 0;			/* make sure i=0 if going through loop second time */
	printf("Enter block number (0 to %lld): ", blocks - 1);
/*****************************************************************/
/* use getchar to get block number so we can check for "\n" only */
/*****************************************************************/
	while (1) {
	    character = (char) getchar();
	    if ((character == '\n') && (i == 0)) {	/* hit return only       */
		/* use last block number */
		printf("Block number = %lld\n", block_number_to_dump);
		for (j = 0; j <= 5; ++j)
		    line[j] = (char) 0x0;
		break;
	    }
	    if (character == '\n') {
		block_number_to_dump = (unsigned int) atoi(line);
		for (j = 0; j <= 5; ++j)
		    line[j] = (char) 0x0;
		break;
	    }
	    line[i] = character;
	    ++i;
	}
	if (block_number_to_dump < 0)
	    break;
	if (block_number_to_dump >= blocks) {
	    printf(" Block number too large for this device.\n");
	    break;
	}
/****************************************************/
/* Dump the block requested so we can look at bytes */
/****************************************************/
	dump(fd, block_number_to_dump, display_mode);
/*  Go get byte number within block to modify  */
	while (1) {
	    printf("Enter byte number to change (1-512): ");
	    if (scanf("%d", &byte) == EOF) {
		printf("scanf error\n");
		exit(EXIT_FAILURE);
	    }
	    if ((0 < byte) && (byte < 513))
		break;
	}
/*  Get new data */
	while (1) {
	    printf("Enter byte (HEX 00-FF):  ");
	    if (scanf("%X", &data_byte) == EOF) {
		printf("scanf error\n");
		exit(EXIT_FAILURE);
	    }
	    if ((0 <= data_byte) && (data_byte <= 255))
		break;
	}

	printf("Block number = %lld\n", block_number_to_dump);

/* Go read the requested block    */

#ifdef _LARGEFILE64_SOURCE
	if ((lseek64(fd, (off64_t) (block_number_to_dump * 512), SEEK_SET)) < 0) {
	    perror("IOtest: lseek64");
	    exit(EXIT_FAILURE);
	}
#else
	if ((lseek(fd, block_number_to_dump * 512, SEEK_SET)) < 0) {
	    perror("IOtest: lseek");
	    exit(EXIT_FAILURE);
	}
#endif
	if ((charsread = (int) read(fd, wbuff, 512)) < 0) {
	    perror("IOtest: read");
	    exit(EXIT_FAILURE);
	}
/********************************************************/
/* See if big_endian flag is set and adjust accordingly */
/********************************************************/
	if (*(char *) &y == 1)
	    big_endian = 0;	/* little endian,  Intel, VAX etc.  */
	else
	    big_endian = 1;	/* big endian,  most UNIX systems   */
	if (big_endian == 0) {	/* little endian? if so, change the byte */
	    wbuff[byte - 1] = (char) data_byte;
/*************************************************************************/
/*  Why [byte-1] in the above line?  It has to do with the byte # from 1 */
/*  to 512 that was requested vs byte 0 to 511 in wbuff[512].....        */
/*************************************************************************/
	} else {		/* big endian, put the new byte in the right place */
	    switch (byte % 4) {
	    case 0:
		wbuff[byte - 4] = (char) data_byte;
		break;
	    case 1:
		wbuff[byte + 2] = (char) data_byte;
		break;
	    case 2:
		wbuff[byte] = (char) data_byte;
		break;
	    case 3:
		wbuff[byte - 2] = (char) data_byte;
		break;
	    }
	}
/*****************************************************************************/
/*  In the above switch, the [byte+/-X]'s adjust for the byte numbering in   */
/*  wbuff[512] (0-511)  vs the 1-512 numbers requested.                      */
/*****************************************************************************/

/*******************************************/
/* write modified block back out to device */
/*******************************************/
#ifdef _LARGEFILE64_SOURCE
	if ((lseek64(fd, (off64_t) (block_number_to_dump * 512), SEEK_SET)) < 0) {
	    perror("IOtest: lseek64");
	    exit(EXIT_FAILURE);
	}
#else
	if ((lseek(fd, block_number_to_dump * 512, SEEK_SET)) < 0) {
	    perror("IOtest: lseek");
	    exit(EXIT_FAILURE);
	}
#endif
	if ((charswrote = (int) write(fd, wbuff, 512)) < 0) {
	    perror("IOtest: write");
	    exit(EXIT_FAILURE);
	}
	dump(fd, block_number_to_dump, display_mode);	/* dump out the changed block  */

	character = (char) getchar();
    }				/* go back and get next block number */
}

/*************************************************************************/
/*DUMP */
static void dump(int fd, long long block_number_to_dump, int display_mode)
/*  Dump the requested block on screen in HEX or ASCII */
/*  if display_mode = 1, dump in hex    */
/*  if display_mode = 0, dump in ascii  */
{
    int ncols, nrows;
    int i, j, jj, k = 0, charsread;
    static unsigned long int rbuf[512 / 4];	/* place to put data */
    short int x1;
    char buff[5];
    char *x;
    unsigned long int mask = 0x000000FFu;	/* mask for first byte */

/*  Seek to block and read 512 bytes    */
#ifdef _LARGEFILE64_SOURCE
    if ((lseek64(fd, (off64_t) (block_number_to_dump * 512), SEEK_SET)) < 0) {
	perror("IOtest: lseek64");
	exit(EXIT_FAILURE);
    }
#else
    if ((lseek(fd, block_number_to_dump * 512, SEEK_SET)) < 0) {
	perror("IOtest: lseek");
	exit(EXIT_FAILURE);
    }
#endif
    if ((charsread = (int) read(fd, rbuf, 512)) < 0) {
	perror("IOtest: read");
	exit(EXIT_FAILURE);
    }
/*  Print Byte numbers for columns and rows on screen   */
    if (display_mode == 1) {
	sprintf(buf, "Byte  1 2 3 4  5 6 7 8  9101112 13141516 17181920 21222324 25262728 29303132 \n");
	ncols = 8;
	nrows = 16;
    } else {
	sprintf(buf, "Byte   1  2  3  4   5  6  7  8   9 10 11 12  13 14 15 16  17 18 19 20 \n");
	ncols = 5;
	nrows = 26;
    }
    printf("%s", buf);
    if (do_log == 1)
	logit(buf);

    sprintf(buf, "----------------------------------------------------------------------------\n  1| ");
    printf("%s", buf);
    if (do_log == 1)
	logit(buf);

/*  i = row number, j = column number, k = 32 bit integer number in block  */
    for (i = 0; i < nrows; ++i) {
	for (j = 0; j < ncols; ++j) {

	    if (i == 25 && j == 3) {
		printf(" <--- byte 512");
		break;		/*  don't finish last row in ascii mode  */
	    }
	    rbuf[k + j] = swap(rbuf[k + j]);
	    if (display_mode == 1) {	/* HEX  */
		sprintf(buf, "%08lX ", (unsigned long int) rbuf[k + j]);
		printf("%s", buf);
		if (do_log == 1)
		    logit(buf);
	    } else {		/*  ASCII  */
		rbuf[k + j] = swap((unsigned long int) rbuf[k + j]);
		for (jj = 1; jj <= 4; ++jj) {
		    x1 = (short int) (rbuf[k + j] & mask);	/*first byte */

		    switch (x1) {
		    case '\0':
			x = " \\0";
			break;
		    case '\007':
			x = " \\a";
			break;
		    case '\b':
			x = " \\b";
			break;
		    case '\f':
			x = " \\f";
			break;
		    case '\n':
			x = " \\n";
			break;
		    case '\r':
			x = " \\r";
			break;
		    case '\t':
			x = " \\t";
			break;
		    case '\v':
			x = " \\v";
			break;
		    default:
			sprintf(buff, "  %c", (char) x1);
			x = (char *) buff;
		    }		/* end switch  */
		    printf("%3s", x);	/* print it as ascii character */
		    rbuf[k + j] = (unsigned int) rbuf[k + j] >> 8;	/* shift the buffer right 8 bits and do it again */

		}		/* end for .. jj  */

		printf(" ");

	    }			/* end else {  */

	}			/* end for ... ncols  */

	k = k + ncols;
	if ((k * 4) < 512)
	    printf("\n%3d| ", k * 4 + 1);
	sprintf(buf, "\n%3d| ", k * 4 + 1);
	if (do_log == 1) {
	    sprintf(buf, "\n%3dl ", k * 4 + 1);
	    logit(buf);
	}
    }				/* end for .. nrows  */

    /*sprintf(buf, "                                                                byte 512 ----^^\n"); */
    /*printf("%s", buf); */
    printf("\n");
}				/* end dump */


static unsigned long int swap(unsigned long int data_to_swap)
{
/***********************************************************/
/* In data_to_swap, bytes are arranged 04030201 This swaps */
/* bytes and returns them as 01020304                      */
/***********************************************************/

    unsigned int x;		/* x marks the spot, it's on the floor to left of your desk */
    unsigned int mask1 = 0x000000FFu;	/* mask for first byte    */
    unsigned int mask2 = 0x0000FF00u;	/* mask for second  byte  */
    x = (mask1 & data_to_swap);
    x = x << 16;
    x = x | (mask2 & data_to_swap);
    x = x << 8;
    data_to_swap = data_to_swap >> 8;
    x = x | (mask2 & data_to_swap);
    data_to_swap = data_to_swap >> 16;
    x = x | (mask1 & data_to_swap);
    return (unsigned long int) (x);
}

/**********************************************************************/
void set_fl(int fd, int flags)
{
    int val;
    if ((val = fcntl(fd, F_GETFL, NULL)) < 0)
	printf("fcntl F_GETFL error");
    val |= flags;
    if (fcntl(fd, F_SETFL, val) < 0)
	printf("fcnt; F_SETFL error");
}

/***********************************************************************/
void flags(int fd)
{
    int accmode, val;
    if ((val = fcntl(fd, F_GETFL, NULL)) < 0)
	printf(" error on fd %d\n", fd);
    accmode = val & O_ACCMODE;
    if (accmode == O_RDONLY)
	printf("read only");
    else if (accmode == O_WRONLY)
	printf("write only");
    else if (accmode == O_RDWR)
	printf("read write");
    else
	printf("unknown access mode");
    if ((val & O_APPEND) == 1)
	printf(", append");
    if ((val & O_NONBLOCK) == 1)
	printf(", nonblocking");
/*#if !defined(_POSIX_SOURCE) && defined(O_SYNC) */
    if ((val & O_SYNC) == 1)
	printf(", sync writes");
/*#endif */
    if (putchar('\n') == EOF) {
	printf("putchar error\n");
	exit(EXIT_FAILURE);
    }
}
