#!/bin/sh
#
# Stresses NFS reads and writes.
#
# Options: none
#
# Requires: cat, cp, diff, mkdir, mount, rm, rmdir,
#           init.d/nfs script
#
. cleanup_functions.sh
# 
# Something's wrong, can't continue.
#
function bail () {
    echo $ERROR_MSG
    cleanup
    exit -2
}
    

#
# For the first test, we are going to nfs mount the local machine's
# /usr/src/linux on the local machine.
# We mount it on the same machine it's exported from to avoid going too
# deep into the IP stack; in particular, we don't want to dip into the
# NIC layer.
#
SOURCE0="/usr/src/linux"
SOURCE1="/usr/src/linux.tmp.$$-s1"
TARGET0="/usr/src/linux.tmp.$$-t0"
TARGET1="/usr/src/linux.tmp.$$-t1"

HOST=127.0.0.1
EXIT_VALUE=0
ERROR_MSG=""

#
# Def's for config files
# 
EXPORTS_FILE="/etc/exports"
EXPORTS_FILE_BACKUP="/etc/exports.original.$$"

#
# Let's make sure $SOURCE0 is a legitimate source.
#
if [  ! -e $SOURCE0 ] ; then {
    ERROR_MSG="$SOURCE0 does not exist."
    ERROR_VALUE=1
    bail
} ; fi

if [ ! -d $SOURCE0 ] ; then {
    ERROR_MSG="$SOURCE0 is not a directory!."
    ERROR_VALUE=2
    bail
} ; fi

#
# Modify exports file
#
if [ ! -f $EXPORTS_FILE ] ; then {
    RESTORE_EXPORTS_FILE="no"
    cat > $EXPORTS_FILE <<EOF
# /etc/exports
#
# Async is the default behaviour.  Does this help stress testing?
#
$SOURCE0	$HOST(rw,no_root_squash,async)
$SOURCE1        $HOST(rw,no_root_squash,async)
EOF
} ; 

else {
    RESTORE_EXPORTS_FILE="yes"
    cp $EXPORTS_FILE $EXPORTS_FILE_BACKUP >>/dev/null 2>&1
    if [ $? -ne 0 ] ; then {
	ERROR_MSG="Can't backup $EXPORTS_FILE."
	ERROR_VALUE=2
	bail
    } ; fi
    cat >> $EXPORTS_FILE <<EOF
#
# Async is the default behaviour.  Does this help stress testing?
#
$SOURCE0	$HOST(rw,no_root_squash,async)
$SOURCE1        $HOST(rw,no_root_squash,async)
EOF
} ; fi

trap "ERROR_VALUE=3; bail" SIGHUP SIGINT SIGTERM

#
# Setup the directories for mounting, exporting
#
mkdir $SOURCE1 >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't make directory $SOURCE1."
    ERROR_VAL=4
    bail
} ; fi

mkdir $TARGET0 >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't make directory $TARGET0."
    ERROR_VAL=4
    bail
} ; fi

mkdir $TARGET1 >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't make directory $TARGET1."
    ERROR_VAL=4
    bail
} ; fi

trap "ERROR_VALUE=4; bail" SIGHUP SIGINT SIGTERM

#
# Check nfs status, start if necessary
#
NFS_STATUS=`/etc/rc.d/init.d/nfs probe` >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't probe NFS status."
    ERROR_VALUE=5
    bail
} ; fi

#
# $NFS_STATUS should be "start" or "reload" here.  If not, there's a bug
# in this script.
#
if [ ! "$NFS_STATUS" ] ; then {
    ERROR_MSG="Bad NFS status NULL."
    ERROR_VALUE=6
    bail
} ; fi
if [ "$NFS_STATUS" != "start" -a "$NFS_STATUS" != "reload" ] ; then {
    ERROR_MSG="Bad NFS status $NFS_STATUS."
    ERROR_VALUE=6
    bail
} ; fi

/etc/rc.d/init.d/nfs $NFS_STATUS >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't modify NFS status using \
	    \"/etc/r.cd/init.d/nfs $NFS_STATUS\"."
    ERROR_VALUE=7
    bail
} ; fi

trap "ERROR_VALUE=8; bail" SIGHUP SIGINT SIGTERM

#
# Mount the NFS exported directories
#
mount -t nfs "$HOST":"$SOURCE0" "$TARGET0" >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't mount $HOST:$SOURCE0 on $TARGET0."
    ERROR_VALUE=8
    bail
} ; fi
mount -t nfs "$HOST":"$SOURCE1" "$TARGET1" >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't mount $HOST:$SOURCE1 on $TARGET1."
    ERROR_VALUE=9
    bail
} ; fi
#
# Blechh.  Broken symlinks causes diff to consider two identical directories
# to be different.  But, it returns 2 (rather than 1) if broken
# symlinks are the only problem (I think.)
#
# Also, maybe checksumming would be a more intelligent way of checking to
# make sure NFS read is idempotent?
#

trap "ERROR_VALUE=10; bail" SIGHUP SIGINT SIGTERM

diff -qr $SOURCE0 $TARGET0 >> /dev/null 2>&1
EXIT_VALUE=$?

if [ $? -eq 2 ] ; then {
    echo "You might have broken symlinks in $SOURCE0." 
#    EXIT_VALUE=2
# Do we want to return an error on this?
#
} ; fi

if [ $? -eq 1 ] ; then {
    ERROR_MSG="There is an error reading from an NFS mounted directory."
    ERROR_VALUE=10
    bail
} ; fi

#
# Push to server from client.
#

#
# Populate TARGET1
#

# 
# We need -d, otherwise cp will bail on broken symlinks.
#
# Should also check just to make sure /usr/src has enough free space.
#

trap "ERROR_VALUE=12; bail" 1 2 5

cp -dr $SOURCE0/* $TARGET1 >> /dev/null 2>&1
#if [ $? -ne 0 ] ; then {
#    ERROR_MSG="Copy to server failed."
#    ERROR_VALUE=11
#    bail
#}

diff -qr $SOURCE0 $TARGET1 >> /dev/null 2>&1
EXIT_VALUE=$?

if [ $? -eq 1 ] ; then {
    ERROR_MSG="Error writing to an NFS mounted directory."
    ERROR_VALUE=12
    bail
} ; fi

trap "" SIGHUP SIGINT SIGTERM

. nfs_kernel_compile 4 $TARGET1

ERROR_VALUE=0
cleanup
exit $EXIT_VALUE

